require_relative 'mirror_manager/version'
require_relative 'mirror_manager/logging'
require_relative 'mirror_manager/sync'
require_relative 'mirror_manager/mirror'
require_relative 'mirror_manager/notification'
require 'English'
require 'yaml'
require 'concurrent'

module Rise
  # Module for managing mirror sync jobs
  module MirrorManager

    def self.run(intervals)
      puts ENV['SLACK_WEBHOOK']
      Manager.execute_jobs(intervals)
    end

    class Manager
      include Rise::MirrorManager::Notification

      LOG = Logging.setup_logger

      def self.execute_jobs(intervals)
        Manager.new(intervals).run_syncs
      end

      def run_syncs()
        thread_pool = Concurrent::FixedThreadPool.new(10, max_queue:50)
        until mirrors.empty?
          mirror = mirrors.shift
          Concurrent::Promises.future(executor: thread_pool) { Sync.run(mirror) }.rescue { |e| send_error_notification(mirror_name: mirror.name, error: e)  }.result
        end
        thread_pool.shutdown
        thread_pool.wait_for_termination
      end

      private

      attr_reader :mirrors, :runtime

      def initialize(intervals)
        @mirrors = retrieve_mirrors(intervals)
        @runtime = Time.now
      end

      def send_error_notification(mirror_name:, error:)
        send_slack_message(mirror_name: mirror_name, error: error, runtime: runtime)
      end

      def retrieve_mirrors(intervals)
        mirrors = []
        intervals.each do |i|
          retrieved_mirrors = Mirror.retrieve_mirrors(Integer(i))
          mirrors.push(*retrieved_mirrors)
        end
        if mirrors.empty?
          raise 'No mirrors were retrieved from #{intervals}.'
        end
        mirrors
      end

      at_exit do
        if $ERROR_INFO
          error = {
            timestamp: Time.now,
            message: $ERROR_INFO.message,
            backtrace: $ERROR_INFO.backtrace,
            gems: Gem.loaded_specs.inject({}) do |m, (n, s)|
            m.merge(n => s.version)
          end
          }
          LOG.error YAML.dump(error)
          send_email YAML.dump(error)
        end
      end
    end
  end
end
