# require File.join(__dir__, '../../../config/initializers/production')
require_relative 'error/cmd_error'
require_relative 'logging'
require_relative 'database'
require_relative 'sync_status'
require 'marz/rsync'
require 'English'
require 'tmpdir'
require 'ostruct'
require 'open3'

module Rise
  module MirrorManager
    # This will create an instance
    # for every mirror rsync job
    class Sync
      include Rise::MirrorManager::Error
      include Rise::MirrorManager::Logging
      include Rise::MirrorManager::SyncStatus
      include Rise::MirrorManager::Database

      DEFAULT_RSYNC_OPTIONS = '-avzuH --no-motd --safe-links --numeric-ids' \
        ' --delay-updates --timeout=600 --contimeout=300'.freeze
      DEFAULT_RSYNC_DELETE_OPTIONS = '--delete --delete-delay'.freeze

      def self.run(mirror)
        Sync.new(mirror).run
      end

      def run
        err = catch :failure do
          if mirror.custom_script
            execute_cmd(mirror.custom_script) { update_sync_status(mirror_name: mirror.name, status: SUCCESS) }
            # execute_custom_script(mirror.custom_script)
          else
            begin
              # err = catch :failure do
              @rsync_opts = build_rsync_opts(mirror.rsync_opts, mirror.rsync_delete_opts)
              # Can't capture the error `File.join(mirror.local_dir, "#{mirror.name}.lock")` when mirror.local_dir == nil
              @lock_file = File.join(mirror.local_dir, "#{mirror.name}.lock")
              establish_lock
              # execute_precmd(mirror.pre_cmd) if mirror.pre_cmd
              execute_cmd(mirror.pre_cmd) if mirror.pre_cmd
              sync
              # execute_postcmd(mirror.post_cmd) if mirror.post_cmd
              execute_cmd(mirror.post_cmd) if mirror.post_cmd
              update_sync_status(mirror_name: mirror.name, status: SUCCESS, total_bandwidth: total_bandwidth)
            ensure
              release_lock
            end
          end
        end
        update_sync_status(mirror_name: mirror.name, status: FAILED, error_msg: err) if err
      end

    private

    def initialize(mirror)
      @mirror = mirror
      @logger = SyncLog.new(mirror.name)
    end


    attr_reader :mirror, :lock_file, :rsync_opts, :logger, :total_bandwidth

    def sync
      Marz::Rsync.run(mirror.remote_source, mirror.local_dir, rsync_opts) do |result|
        if result.success?
          @total_bandwidth = convert_bytes(result.total_bytes_sent + result.total_bytes_received)
          logger.log_info result.output
        else
          logger.log_error result.output
          throw :failure, result.output
        end
      end
    end

    def execute_cmd(cmd)
      Open3.popen2e(cmd.to_s) do |_i, oe, t|
        unless t.value.success?
          throw :failure, oe.read
        end
      end
      yield if block_given?
    end

    # def execute_precmd(cmd)
    #   Open3.popen2e(cmd.to_s) do |_i, oe, t|
    #     unless t.value.success?
    #       update_precmd_result(mirror.name, FAILED)
    #       msg = oe.read # Can only be retrieved/read once
    #       logger.log_error(msg) unless msg.nil? || msg.empty?
    #       exit
    #     end
    #     update_precmd_result(mirror.name, SUCCESS)
    #   end
    # rescue Errno::ENOENT => e
    #   update_precmd_result(mirror.name, FAILED)
    #   logger.log_error(e.message) unless e.message.nil? || e.message.empty?
    #   exit
    # end

    # def execute_postcmd(cmd)
    #   Open3.popen2e(cmd.to_s) do |_i, oe, t|
    #     unless t.value.success?
    #       update_postcmd_result(mirror.name, FAILED)
    #       msg = oe.read # Can only be retrieved/read once
    #       logger.log_error(msg) unless msg.nil? || msg.empty?
    #       exit
    #     end
    #     update_postcmd_result(mirror.name, SUCCESS)
    #   end
    # rescue Errno::ENOENT => e
    #   update_postcmd_result(mirror.name, FAILED)
    #   logger.log_error(e.message) unless e.message.nil? || e.message.empty?
    #   exit
    # end

    # def execute_custom_script(cmd)
    #   execute_cmd(cmd) do |_oe, t|
    #     update_sync_status(mirror.name, SUCCESS) if t.value.success?
    #   end
    # end

    # def execute_cmd(cmd)
    #   Open3.popen2e(cmd.to_s) do |_i, oe, t|
    #     yield(oe, t) if block_given?
    #     process_sync_failure(oe.read) unless t.value.success?
    #   end
    # rescue Errno::ENOENT => e
    #   process_sync_failure(e.message)
    # end

    def process_sync_failure(msg = nil)
      update_sync_status(mirror.name, FAILED)
      logger.log_error(msg) unless msg.nil? || msg.empty?
      exit
    end

    def establish_lock
      if File.exist?(lock_file)
        acquire_lock
      else
        create_lock
      end
    end

    def acquire_lock
      pid = File.read(lock_file).to_i
      Process.kill(0, pid)
    rescue Errno::ESRCH
      File.delete(lock_file) # nonexistent process
      create_lock
    rescue Errno::EPERM
      throw :failure, "No permission/privilege to send signal to process(#{pid})"
    else
      throw :failure, "Cannot establish lock. Another process(#{pid}) is currently running and holding the lock."
    end

    def create_lock
      File.open(lock_file, File::RDWR | File::CREAT, 0o644) do |f|
        f.flock(File::LOCK_EX)
        f.write(Process.pid.to_s)
        f.flush
      end
    end

    def release_lock
      File.delete(lock_file) if File.exist?(lock_file)
    end

    def build_rsync_opts(rsync_opts, rsync_delete_opts)
      opts = rsync_opts || DEFAULT_RSYNC_OPTIONS
      opts += ' ' + (rsync_delete_opts || DEFAULT_RSYNC_DELETE_OPTIONS)
      opts.split(' ')
    end

    def convert_bytes(bytes)
      bytes = Integer(bytes)
      result =
        if bytes <= 104_857_0
          format('%.2fK', (bytes / 1024.0))
        elsif bytes <= 107_373_645_5 # 0.99GB
          format('%.2fM', ((bytes / 1024.0) / 1024.0))
        else
          format('%.2fG', (((bytes / 1024.0) / 1024.0) / 1024.0))
        end
      result.gsub(/(\.)0+/, '')
    end

  end
end
end

# mirror = OpenStruct.new(name:"Delete Me", remote_source:"/tmp/srv/mirror/delete-me-source", local_dir:"/tmp/srv/mirror/delete-me-dest", rsync_opts:'-avzuH --no-motd --safe-links --numeric-ids --delay-updates --timeout=600', rsync_del_opts:nil, exclude:nil, include:nil, filter:nil, precmd:nil, postcmd:nil, depth:1)
# Rise::MirrorManager::Sync.new(mirror).run
